import argparse
import inspect
import logging

from atils.common import config

logging.basicConfig(level=config.get_logging_level())  # type: ignore


# ==========================================
# Stuff for actually running test functions
# ==========================================
def main(args: str):
  parser: argparse.ArgumentParser = argparse.ArgumentParser()

  parser.add_argument(
    "action",
    help="Select which action to perform. Defaults to run",
    default="run",
    nargs="?",
  )

  parser.add_argument("--function-name", type=str)

  arguments: argparse.Namespace = parser.parse_args(args)

  if arguments.action == "run":
    if arguments.function_name is None:
      logging.error("No function name specified")
      exit(1)
    else:
      function = globals().get(arguments.function_name)
      if function is None:
        logging.error(f"Function {arguments.function_name} not found")
        exit(1)
      else:
        function()
  elif arguments.action == "list":
    functions = _get_function_list()
    if len(functions) == 0:
      print("There are no test functions here at the moment")
      exit(0)
    for function in functions:
      print(function)
  else:
    logging.error(
      "Invalid action. If you are trying to run a specific function, remember to use --function-name",
    )


def _get_function_list():
  function_list = []
  for name, obj in inspect.getmembers(inspect.getmodule(inspect.currentframe())):
    if inspect.isfunction(obj) and not name.startswith("_") and name != "main":
      function_list.append(name)
  return function_list


# =======================
# Functions being tested
# =======================



from rich.align import Align
from rich.console import Console
from rich.table import Table


def print_jobconfig(jobconfig_data: dict) -> None:
  """
  Given a jobconfig dictionary, print it to the console as a table

  Args:
    jobconfig_data (dict): The jobconfig dictionary representation
  """
  console = Console()

  table = Table(show_header=True, header_style="bold", show_lines=True)
  table.add_column(jobconfig_data["display_name"], justify="center")
  table.add_row(jobconfig_data["long_description"])

  args_subtable = Table(show_header=True)
  table.add_row("Arguments", style="bold")
  if "args" in jobconfig_data.keys() and jobconfig_data["args"] is not None:
    args_subtable.add_column("Name")
    args_subtable.add_column("Description")
    args_subtable.add_column("Required")
    args_subtable.add_column("Options")

    for arg in jobconfig_data["args"]:
      args_subtable = _add_row_to_subtable(args_subtable, arg)
    table.add_row(Align.center(args_subtable))
  else:
    table.add_row("Good news! This job has no arguments")

  console.print(table)


def _add_row_to_subtable(subtable: Table, row: dict) -> Table:
  """
  Takes a Table representing the args subtable and returns a Table with the row added to it
  This function accounts for missing items so they're not printed as empty

  Args:
    subtable (Table): The Table representing the args subtable
    row (dict): A dictionary representing a single argument

  Returns:
    Table: The Table with the row added to it
  """
  options_string = " ".join(row["options"])

  if "description" not in row.keys():
    row["description"] = "N/A"

  subtable.add_row(
    row["name"], row["description"], str(row["required"]), options_string
  )
  return subtable



import logging
import os

from atils.common.settings import settings


def get_full_atils_dir(dir_name: str) -> str:
  """
  Given the name of a subdirectory, return the result of using os.join
  on that directory and the directory where our monorepo is installed
  Args:
    dir_name (str): The name of the subdirectory to join
  Returns:
    str: The full path to the subdirectory
  """
  return os.path.join(settings.INSTALL_DIR, settings[dir_name])


def get_logging_level() -> int:
  """
  Return the logging level configured. We use this function so we return the
  constants in the logging module, rather than a string we need to parse anyway
  """
  if settings.LOG_LEVEL == "DEBUG":
    return logging.DEBUG
  elif settings.LOG_LEVEL == "INFO":
    return logging.INFO
  elif settings.LOG_LEVEL == "WARNING":
    return logging.WARNING
  elif settings.LOG_LEVEL == "ERROR":
    return logging.ERROR
  elif settings.LOG_LEVEL == "CRITICAL":
    return logging.CRITICAL
  else:
    logging.warning("No log level set, defaulting to INFO")
    return logging.INFO



import argparse
import base64
import logging
import os
import shutil
import sys

from atils.common import config, yaml_utils
from atils.common.settings import settings
from kubernetes import client
from kubernetes import config as k8s_config

client.rest.logger.setLevel(logging.WARNING)

logging.basicConfig(level=config.get_logging_level())  # type: ignore


def main(args: list[str]):
  # This variable tracks whether or not we have configuration available to run kubernetes commands
  CAN_RUN: bool = load_config()

  if not CAN_RUN:
    logging.error("No configuration available to run kubernetes commands")
    exit(1)

  # TODO add autocomplete for secret names
  parser: argparse.ArgumentParser = argparse.ArgumentParser()
  subparsers = parser.add_subparsers(
    help="Select a subcommand",
    dest="subparser_name",
  )

  # Options for managing kubernetes secrets. This is different from Vault secrets management
  secrets_parser = subparsers.add_parser(
    "secrets",
    help="Commands to manage kubernetes secrets",
  )

  secrets_parser.add_argument(
    "command",
    choices=["decode"],
    help="Which command to use to operate on secrets",
  )

  secrets_parser.add_argument(
    "secret_name",
    help="The name of the secret to operate on",
  )

  secrets_parser.add_argument(
    "-n",
    "--namespace",
    help="The namespace of the secret to operate on",
  )

  istio_parser = subparsers.add_parser(
    "istio",
    help="Commands to manage or interact with istio",
  )
  istio_parser.add_argument(
    "command",
    choices=["label-namespaces"],
    help="Which istio-related operator to perform",
  )

  if len(args) == 0:
    parser.print_help(sys.stderr)
    sys.exit(1)

  arguments: argparse.Namespace = parser.parse_args(args)
  args_dict: dict[str, str] = vars(arguments)

  if arguments.subparser_name == "secrets":
    if args_dict["command"] == "decode":
      if args_dict["secret_name"] is None:
        logging.error("Error: A secret name must be provided ")
        sys.exit(1)
      else:
        get_and_decode_secret(args_dict["secret_name"], args_dict["namespace"])

  elif arguments.subparser_name == "istio":
    if args_dict["command"] == "label-namespaces":
      _label_namespaces_for_istio_injection(
        [
          "ingress-nginx",
          "istio-system",
          "kube-public",
          "kube-system",
          "local-path-storage",
        ],
      )

      _clear_pods_for_istio_injection(
        [
          "ingress-nginx",
          "istio-system",
          "kube-public",
          "kube-system",
          "local-path-storage",
        ],
      )


def check_namespace_exists(namespace_name: str) -> bool:
  """Check if a given kubernetes namespace exists
  Args:
    namespace_name (str): The name of the namespace to check
  Returns:
    bool: True if the namespace exists, False otherwise.
  """
  # Create Kubernetes API client
  v1 = client.CoreV1Api()

  # Use the API client to list all namespaces
  namespace_list = v1.list_namespace().items

  # Check if the namespace exists
  if any(namespace.metadata.name == namespace_name for namespace in namespace_list):
    return True
  else:
    return False


def get_and_decode_secret(secret_name: str, secret_namespace: str) -> None:
  """Get a kubernetes secret, and then decode and pretty print it
  Args:
    secret_name (str): The name of the secret to decode and print
    secret_namespace (str): The namespace the given secret is located in
  """
  try:
    # Create a Kubernetes API client
    api = client.CoreV1Api()

    if secret_namespace is None:
      current_context = k8s_config.list_kube_config_contexts()[1]
      # Check if current_context["context"]["namespace"] is None
      secret_namespace = current_context.get("context").get(
        "namespace",
        "default",
      )

    # Get the Secret from the specified namespace
    secret = api.read_namespaced_secret(
      name=secret_name,
      namespace=secret_namespace,
    )

    terminal_width = shutil.get_terminal_size().columns

    print("=" * int(terminal_width / 2))
    print(secret.metadata.name.center(int(terminal_width / 2)))
    print("=" * int(terminal_width / 2))

    # Decode and pretty print the data items
    for key, value in secret.data.items():
      decoded_value = base64.b64decode(value).decode("utf-8")
      # If decoded_value is more than one line, print on a new line
      if "\n" in decoded_value:
        print(f"{key}:")
        print(decoded_value)
      else:
        print(f"{key}: {decoded_value}")

      print("=" * int(terminal_width / 2))

  except Exception as e:
    logging.exception(f"An error occurred: {e}")


def get_current_namespace() -> str:
  """Gets the current namespace set in the current context
  Returns:
    str: The current namespace set in the current context
  """
  contexts, active_context = k8s_config.list_kube_config_contexts()

  if "namespace" in active_context["context"].keys():
    return active_context["context"]["namespace"]
  else:
    return "default"


def load_config() -> bool:
  """Loads some kind of kubernetes configuration, either in-cluster or in kubeconfig,
  and returns true if it is able to. If it cannot, it will return false, and the main
  function is in charge of failing if it needs to

  Returns
  -------
    (bool): True if a config could be loaded, false otherwise

  """
  # TODO also support running from a pod
  if "KUBECONFIG" in os.environ:
    k8s_config.load_kube_config(os.environ["KUBECONFIG"])
    return True
  # Check if ~/.kube/config exists
  if os.path.exists(os.path.expanduser("~/.kube/config")):
    k8s_config.load_kube_config()
    return True
  else:
    logging.info(
      "Could not find a kubeconfig. Any commands related to kubernetes have been disabled",
    )
    return False

def _clear_pods_for_istio_injection(excluded_namespaces: list[str]) -> None:
  try:
    # Create a Kubernetes API client
    api = client.CoreV1Api()

    # List all namespaces
    namespaces = api.list_namespace().items

    # Iterate over each namespace
    for namespace in namespaces:
      namespace_name = namespace.metadata.name

      # Check if the namespace is in the excluded list
      if namespace_name not in excluded_namespaces:
        # List all pods in the namespace
        pods = api.list_namespaced_pod(namespace_name).items

        # Iterate over each pod
        for pod in pods:
          # Check if the pod already has an Istio sidecar injected
          if "istio-proxy" in [container.name for container in pod.spec.containers]:
            print(
              f"Skipping pod '{pod.metadata.name}' in namespace '{namespace_name}' (Istio sidecar already"
              + " injected).",
            )
            continue

          # Check if the pod has the annotation sidecar.istio.io/inject: "false"
          if pod.metadata.annotations and pod.metadata.annotations.get("sidecar.istio.io/inject") == "false":
            print(
              f"Skipping pod '{pod.metadata.name}' in namespace '{namespace_name}' (annotation"
              + ' sidecar.istio.io/inject: "false").',
            )
            continue

          # Delete the pod
          api.delete_namespaced_pod(pod.metadata.name, namespace_name)
          print(
            f"Deleted pod '{pod.metadata.name}' in namespace '{namespace_name}' for Istio injection.",
          )

  except Exception as e:
    print(f"Error: {e!s}")


def _label_namespaces_for_istio_injection(excluded_namespaces: list[str]) -> None:
  try:
    # Create a Kubernetes API client
    api = client.CoreV1Api()

    # List all namespaces
    namespaces = api.list_namespace().items

    # Iterate over each namespace
    for namespace in namespaces:
      namespace_name = namespace.metadata.name

      # Check if the namespace is in the excluded list
      if namespace_name not in excluded_namespaces:
        # Add or update the Istio injection label
        labels = namespace.metadata.labels or {}
        labels["istio-injection"] = "enabled"
        namespace.metadata.labels = labels

        # Update the namespace
        api.patch_namespace(namespace_name, namespace)
        print(f"Labeled namespace '{namespace_name}' for Istio injection.")
      else:
        print(f"Skipping namespace '{namespace_name}' (excluded).")

  except Exception as e:
    print(f"Error: {e!s}")



import os

from jinja2 import Environment, FileSystemLoader, FunctionLoader
from pkg_resources import resource_string


def load_from_self(filename: str) -> str:
  """
  Given the name of a template file in the atila/templates directory, load the
  contents of the file as a bytes object

  Args:
    filename (str): The name of the file to load

  Returns:
    bytes: The contents of the file as a bytes object
  """
  return resource_string("atils.templates", filename).decode("utf-8")


def template_file(template_path: str, args_dict: dict) -> str:
  """
  Given the path to a template file, load the template and substitute the
  values in the dictionary into the template.

  Args:
    template_path (str): The path to the template file
    args_dict (dict): A dictionary of values to substitute into the template

  Returns:
    str: The substituted template content as a string.
  """
  env = Environment(loader=FunctionLoader(load_from_self))
  template = env.get_template(os.path.basename(template_path))
  substituted_content = template.render(args_dict)

  return substituted_content


def template_external_file(template_path: str, args_dict: dict) -> str:
  """
  Given the path to a template file that is not bundled with atils,load
  the template and substitute the values in the dictionary into the template.

  Args:
    template_path (str): The path to the template file
    args_dict (dict): A dictionary of values to substitute into the template

  Returns:
    str: The substituted template content as a string.
  """
  env = Environment(loader=FileSystemLoader(os.path.dirname(template_path)))
  template = env.get_template(os.path.basename(template_path))
  substituted_content = template.render(args_dict)

  return substituted_content


def template_file_and_output(
  template_path: str, output_path: str, args_dict: dict
) -> None:
  """
  Given a path to a template file, and arguments, template the file and then output
  it to a file

  Args:
    template_path (str): The path to the template file
    output_path (str): The path to the output file
    args_dict (dict, optional): A dictionary of values to substitute into the template. Defaults to None.
  """
  substituted_content = template_file(template_path, args_dict)

  with open(output_path, "w") as output_file:
    output_file.write(substituted_content)



import argparse
import json
import logging
import os
import subprocess

from atils.common import config

logging.basicConfig(level=config.get_logging_level())  # type: ignore


def main(args: list[str]):
  parser: argparse.ArgumentParser = argparse.ArgumentParser()

  parser.add_argument(
    "action",
    help="Select which action to perform. Defaults to build",
    default="build",
    nargs="?",
  )

  parser.add_argument(
    "--build-directory",
    "-bd",
    type=str,
    help="The directory where .atils_buildconfig.json is located",
  )

  parser.add_argument(
    "--actions-only",
    "-ao",
    help="Only show actions",
    action="store_true",
  )
  parser.add_argument(
    "--action-sets-only",
    "-aso",
    help="Only show action sets",
    action="store_true",
  )

  parser.add_argument(
    "--actions",
    nargs="*",
    help="Actions defined in .atils_buildconfig.json of current directory to run",
  )

  parser.add_argument(
    "--action-set",
    help="A single action set defined in .atils_buildconfig.json of current directory to run",
  )

  arguments: argparse.Namespace = parser.parse_args(args)

  directory: str = os.getcwd()
  show_actions: bool = True
  show_action_sets: bool = True

  if arguments.actions_only:
    show_actions = False
  if arguments.action_sets_only:
    show_action_sets = False

  if not show_actions and not show_action_sets:
    raise ValueError("You need to allow showing either actions or action sets")

  if arguments.build_directory is not None:
    directory = arguments.build_directory

  if arguments.action == "list":
    run_list_action(directory, show_actions, show_action_sets)
  elif arguments.actions is not None:
    run_build_actions(arguments.actions, directory)
  elif arguments.action_set is not None:
    run_build_action_set(arguments.action_set, directory)
  else:
    default_action_set: str = _get_default_action_set(directory)
    if default_action_set:
      run_build_action_set(default_action_set, directory)
    else:
      logging.error(
        "No default action set defined. Either define one in"
        + f" {os.path.join(directory, '.atils_buildconfig.json')} or use an argument to specify what to do",
      )


def validate_listed_actions(
  available_actions: list[dict],
  listed_actions: list[str],
) -> None:
  """Take a list of valid actions, a list of actions provided by a user, and ensures that all
  actions provided by the user are valid.

  Args:
  ----
    available_actions (list[object]): A list of objects representing available actions.
    listed_actions (list[str]): A list of strings representing actions provided by the user.

  Raises:
  ------
    ValueError: If any action provided by the user is not valid.

  Returns:
  -------
    None.

  """
  available_action_names = [action["name"] for action in available_actions]
  for action in listed_actions:
    if action not in available_action_names and action != "all":
      raise ValueError(f"{action} is not a valid action")


def run_list_action(directory: str, show_actions: bool, show_action_sets: bool) -> None:
  """List all available build actions, or only actions or only action sets.

  Args:
  ----
    directory (str): The directory where our .atils_buildconfig.json file to list is located
    show_actions (bool): Whether to show actions or not
    show_action_sets (bool): Whether to show action sets or not
  Raises:
    ValueError: If both show_actions and show_action_sets are True.

  Returns:
  -------
    None

  """
  available_actions = _get_available_actions(directory)
  available_action_sets = _get_available_action_sets(directory)

  if not show_actions and not show_action_sets:
    raise ValueError("Cannot specify both actions_only and action_sets_only")
  if show_actions:
    print("Actions")
    print("=======================================")
    for action in available_actions:
      _print_action(action)
  if show_action_sets:
    print("Action Sets")
    print("=======================================")
    for action_set in available_action_sets:
      _print_action_set(action_set, available_actions)


def run_build_action_set(action_set_name: str, directory: str) -> None:
  """Runs all the actions defined in a given action_set
  Args:
    action_set (dict): A dictionary representing an action set
    directory (str): The directory where our .atils_buildconfig.json file to list is located
  """
  available_actions: list[dict] = _get_available_actions(directory)
  action_set: dict = _get_action_set(action_set_name, directory)

  validate_listed_actions(available_actions, action_set["actions"])

  if "strict" in action_set and not action_set["strict"]:
    for action in available_actions:
      if action["name"] in action_set["actions"]:
        _run_action(action, directory)
  else:
    for action in available_actions:
      if action["name"] in action_set["actions"]:
        _run_action_strict(action, directory)


def run_build_actions(listed_actions: list[str], directory: str) -> None:
  # TODO allow us to mark if all build actions must pass
  """Run user-specified build actions, in the order specified by .atils_buildconfig.json.

  Args:
  ----
    listed_actions (list[str]): The list of actions provided by the user.
    directory (str): The directory to use as the current working directory.

  """
  available_actions = _get_available_actions(directory)

  validate_listed_actions(available_actions, listed_actions)

  if "all" in listed_actions:
    for action in available_actions:
      _run_action(action, directory)
  else:
    # Since I know you might ask... we do it this way to ensure that the actions are run in order.
    # The user can enter them arbitrarily, but available_actions is sorted, so we can use that as
    # our source of truth
    for action in available_actions:
      if action["name"] in listed_actions:
        _run_action(action, directory)


def _get_available_actions(directory: str) -> list[dict]:
  """List all actions available in a .atils_buildconfig.json file in the given directory.
  The file must be in the given directory and must be a valid JSON file.
  The file must contain a list of actions, each action must be an object with the following keys:
    - name: The name of the action
    - command: The command to run
    - order: The order in which to run the action. Lower numbers are run first.
  The actions are returned sorted by their order.

  Arguments:
  ---------
    directory (str): The directory where the .atils_buildconfig.json file is located.

  Returns:
  -------
    A list of objects representing actions available in the .atils_buildconfig.json file.

  """
  filename: str = os.path.join(directory, ".atils_buildconfig.json")

  if os.path.isfile(filename):
    with open(filename) as f:
      data = json.load(f)
      if "actions" not in data:
        raise ValueError(f"No actions found in {filename}")
      return sorted(data["actions"], key=lambda x: x["order"])
  else:
    raise FileNotFoundError(f"{filename} does not exist")


def _get_available_action_sets(directory: str) -> list[dict]:
  """List all action sets available in a .atils_buildconfig.json file in the given directory.
  The file must be in the given directory and must be a valid JSON file.
  The file must contain a list of action sets, each action set must be an object with the following keys:
    - name: The name of the action set
    - actions: A list of actions in the action set
  The action sets are returned sorted by their name.

  Arguments:
  ---------
    directory (str): The directory where the .atils_buildconfig.json file is located.

  Returns:
  -------
    A list of objects representing action sets available in the .atils_buildconfig.json file. Returns an empty
    list if no such list is found

  """
  filename = os.path.join(directory, ".atils_buildconfig.json")

  if os.path.isfile(filename):
    with open(filename) as f:
      data = json.load(f)
      if "action_sets" not in data:
        return []
      return sorted(data["action_sets"], key=lambda x: x["name"])
  else:
    raise FileNotFoundError(f"{filename} does not exist")


def _get_action_set(action_set: str, directory: str) -> dict:
  """Get a given action set from an .atils_buildconfig.json file
  Args:
    action_set (str): The name of the action set to get
    directory (str): The directory where the .atils_buildconfig.json file is located
  Returns:
    A dictionary representing the action set, or an empty dictionary if the action set is not found
  """
  available_action_sets = _get_available_action_sets(directory)

  for action_set_obj in available_action_sets:
    if action_set_obj["name"] == action_set:
      return action_set_obj
  raise ValueError(
    f"Action set {action_set} not available in {os.path.join(directory, '.atils_buildconfig.json')}",
  )


def _get_default_action_set(directory: str) -> str:
  """Given a directory with an .atils_buildconfig.json file, return the name of the default action set.

  Args:
  ----
    directory (str): The directory where the .atils_buildconfig.json file is located
  Returns:
    A string representing the name of the default action set, or an empty string if no default action set is found.

  """
  available_action_sets: list[dict] = _get_available_action_sets(directory)
  for action_set in available_action_sets:
    if action_set.get("default"):
      return action_set["name"]
  return ""


def _print_action(action: dict) -> None:
  """Print the name and command of an action.

  Args:
  ----
    action (object): An object representing an action from a .atils_buildconfig.json file.

  """
  if "description" in action:
    print(f"{action['name']} | {action['description']}")
  else:
    print(f"{action['name']} | {action['command']}")
  print()


def _print_action_set(action_set: dict, actions: list[dict]) -> None:
  """Print the name and actions of an action set.

  Args:
  ----
    action_set (dict): An object representing an action set from a .atils_buildconfig.json file.
    actions (list[dict]): A list of objects, representing all available actions

  """
  print(f"{action_set['name']}:")
  if "description" in action_set:
    print(f"{action_set['description']}")
  for action in action_set["actions"]:
    for a in actions:
      if a["name"] == action:
        action_dict = a
        break
    if "description" in action:
      print(f"  {action} | {action_dict['description']}")
    else:
      print(f"  {action} | {action_dict['command']}")
  print()


def _run_action(action: dict, directory: str) -> None:
  """Run the command from an action, using subprocess.run
  Args:
    action (object): An object representing an action.
  """
  _validate_action_can_run(action)
  if action["command"] == "":
    logging.error("An atils build command can't be blank")
    exit(1)
  subprocess.run(action["command"], shell=True, cwd=directory, check=False)


def _run_action_strict(action: dict, directory: str) -> None:
  """Run the command from an action, using subprocess.run. Fails if the command fails
  Args:
    action (object): An object representing an action
  """
  _validate_action_can_run(action)
  if action["command"] == "":
    logging.error("An atils build command can't be blank")
    exit(1)
  try:
    subprocess.run(action["command"], shell=True, cwd=directory, check=True)
  except subprocess.CalledProcessError:
    logging.exception(f"Error running {action['name']}")
    exit(1)


def _validate_action_can_run(action: dict):
  """Given an action, checks if it is a CI-only action. If it is, and we are not running in a CI environment,
  exit the program.

  Args:
  ----
    action (dict): A dict representing an action from a .atils_buildconfig.json file.

  Returns:
  -------
    None.

  Raises:
  ------
    SystemExit: If we are not running in a CI environment and the action is a CI-only action.

  """  # noqa: D205
  if "ci_only" in action and action["ci_only"] and "ATILS_CI_ENV" not in os.environ:
    logging.error(
      f"Attempted to run a CI only-action ({action['name']}) in a non-CI environment. "
      + "If you know what you're doing, set ATILS_CI_ENV to true.",
    )
    exit(1)



import sys

from atils import argocd
from atils import atils_kubernetes as kubernetes
from atils import build, helm, jobs, test


def main():
  if len(sys.argv) < 2:
    print("Atils requires at least one subcommand argument.")
    sys.exit(1)

  script_name: str = sys.argv[1]

  if script_name == "kubernetes":
    kubernetes.main(sys.argv[2:])
  elif script_name == "argocd":
    argocd.main(sys.argv[2:])
  elif script_name == "build":
    build.main(sys.argv[2:])
  elif script_name == "job" or script_name == "jobs":
    jobs.main(sys.argv[2:])
  elif script_name == "helm":
    helm.main(sys.argv[2:])
  elif script_name == "test":
    test.main(sys.argv[2:])
  elif script_name == "help" or script_name == "--help":
    print(
      "Collection of utility scripts. For more information, call one of the subcommands with '--help'"
    )
    print("Usage: atils <subcommand> [options]")
    print("Subcommands:")
    print("  kubernetes: Manage Kubernetes resources")
    print("  argocd: Manage Argo CD resources")
    print("  build: Manage build resources")
    print("  job: Manage job resources")
    print("  helm: Manage Helm resources")
    sys.exit(0)
  else:
    print(f"Unrecognized subcommand: {script_name}")
    print("Valid subcommands are: kubernetes, argocd, build, job, helm")
    sys.exit(1)



import argparse
import itertools
import json
import logging
import os
import sys
import termios
import time
import tty
from threading import Thread
from typing import Any, List

import yaml
from kubernetes.client.rest import ApiException
from kubernetes.stream import stream

from atils import atils_kubernetes
from atils.common import config, console_utils, template_utils
from kubernetes import client
from kubernetes import config as k8s_config
from kubernetes import utils

client.rest.logger.setLevel(logging.ERROR)

logging.basicConfig(level=config.get_logging_level())  # type: ignore


def main(args: str) -> None:
  # This variable tracks whether or not we have configuration available to run kubernetes commands
  CAN_RUN: bool = atils_kubernetes.load_config()

  if not CAN_RUN:
    logging.error("No configuration available to run kubernetes commands")
    exit(1)

  parser: argparse.ArgumentParser = argparse.ArgumentParser()
  subparsers = parser.add_subparsers(
    help="Commands to manage kubernetes jobs",
    dest="subparser_name",
  )

  run_parser = subparsers.add_parser("run")
  run_parser.add_argument("job_name", help="Name of the job to run")
  # TODO Add some values for jobs, if needed, so we can set them with this argument
  run_parser.add_argument(
    "--set",
    help="Set values to fill in job template. WIP, not currently working",
    nargs=2,
    action="append",
    dest="values",
  )

  pvc_parser = subparsers.add_parser("manage-pvc")
  pvc_parser.add_argument(
    "--pvc-name",
    "-pn",
    help="The name of the PVC to launch a management pod for",
  )
  pvc_parser.add_argument(
    "--namespace",
    "-n",
    help="The namespace the PVC is located in. Defaults to current namespace",
  )

  postgres_parser = subparsers.add_parser("postgres-manager")

  devterm_parser = subparsers.add_parser("devterm")
  devterm_parser.add_argument(
    "--namespace",
    "-n",
    help="The namespace to launch the devterm in. Defaults to current namespace",
  )

  list_parser = subparsers.add_parser("list")

  describe_parser = subparsers.add_parser("describe")
  describe_parser.add_argument("job_name", help="Then name of the job to describe")

  arguments: argparse.Namespace = parser.parse_args(args)

  if arguments.subparser_name == "run":
    args_dict = vars(arguments)
    job_args = {}

    if args_dict["values"] is not None:
      for arg in args_dict["values"]:
        job_args[arg[0]] = arg[1]

    jobconfig_args = _get_arguments_from_jobconfig(args_dict["job_name"])

    args_filled_from_command_lines = _fill_out_jobconfig_args_from_command_line(
      jobconfig_args,
      job_args,
    )

    if "" in args_filled_from_command_lines.values():
      final_job_args = _get_missing_arguments_interactive(
        args_filled_from_command_lines,
      )
    else:
      final_job_args = args_filled_from_command_lines

    run_job(arguments.job_name, final_job_args)
  elif arguments.subparser_name == "manage-pvc":
    args_dict = vars(arguments)
    current_namespace = atils_kubernetes.get_current_namespace()

    if "namespace" in args_dict.keys():
      if args_dict.get("namespace") is not None:
        launch_pvc_manager(args_dict["pvc_name"], args_dict["namespace"])
      else:
        launch_pvc_manager(args_dict["pvc_name"], current_namespace)
    else:
      launch_pvc_manager(args_dict["pvc_name"], current_namespace)
  elif arguments.subparser_name == "postgres-manager":
    launch_postgres_manager()
  elif arguments.subparser_name == "devterm":
    namespace: str = ""
    args_dict: dict = vars(arguments)

    if args_dict["namespace"] is not None:
      namespace = args_dict["namespace"]
    else:
      namespace = atils_kubernetes.get_current_namespace()

    launch_devterm_pod(namespace)
  elif arguments.subparser_name == "list":
    list_available_jobs()
  elif arguments.subparser_name == "describe":
    args_dict = vars(arguments)
    describe_job(args_dict["job_name"])
  else:
    logging.error(f"Unrecognized command {arguments.subparser_name}")
    exit(1)


def describe_job(job_name: str) -> None:
  """Using the .atils_jobconfig.json file, describe a job's purpose, as well as any arguments it takes.

  Args:
  ----
    job_name (str): The name of the job to describe

  """
  jobs_dir = config.get_full_atils_dir("JOBS_DIR")
  dir_for_job = os.path.join(jobs_dir, job_name)

  if os.path.exists(dir_for_job):
    jobconfig_path = os.path.join(dir_for_job, ".atils_jobconfig.json")

    if os.path.exists(jobconfig_path):
      with open(jobconfig_path) as file:
        jobconfig_data = json.load(file)
        console_utils.print_jobconfig(jobconfig_data)
    else:
      logging.error(
        f"Job {job_name} does not have a jobconfig. It probably will not run successfully",
      )
      exit(1)
  else:
    logging.error(f"Job {job_name} does not exist")
    exit(1)


def launch_postgres_manager() -> None:
  """Launch a postgres client pod, connected to the master user"""
  pod_manifest: dict[str, Any] = {
    "apiVersion": "v1",
    "kind": "Pod",
    "metadata": {"name": "postgres-manager"},
    "spec": {
      "containers": [
        {
          "name": "postgres-manager",
          "image": "aidanhilt/atils-postgres-client",
          "command": ["sh", "-c", "sleep 1800"],
          "env": [
            {
              "name": "PGHOST",
              "value": "postgres-postgresql.postgres.svc.cluster.local",
            },
            {"name": "PGUSER", "value": "postgres"},
            {
              "name": "PGPASSWORD",
              "valueFrom": {
                "secretKeyRef": {
                  "name": "postgres-config",
                  "key": "postgres-password",
                },
              },
            },
            {"name": "PGDATABASE", "value": "postgres"},
          ],
        },
      ],
      "restartPolicy": "Never",
    },
  }
  _create_pod_if_not_existent("postgres-manager", pod_manifest, "postgres")
  _exec_shell_in_pod(
    "postgres-manager",
    "postgres",
    "postgres-manager",
    ["psql"],
    True,
  )


def launch_devterm_pod(namespace: str) -> None:
  """Launch a pod with assorted devterm utils"""
  pod_manifest: dict[str, Any] = {
    "apiVersion": "v1",
    "kind": "Pod",
    "metadata": {"name": "atils-devterm"},
    "spec": {
      "serviceAccountName": "debug-job",
      "containers": [
        {
          "name": "atils-devterm",
          "image": "aidanhilt/atils-debug",
          "command": ["sh", "-c", "sleep 1800"],
        },
      ],
      "restartPolicy": "Never",
    },
  }
  _ensure_access_exists("debug-job", namespace, "admin", True)
  _create_pod_if_not_existent("atils-devterm", pod_manifest, namespace)
  _exec_shell_in_pod("atils-devterm", namespace, "atils-devterm", delete=True)


def launch_pvc_manager(pvc_name: str, namespace: str) -> None:
  """Given the name of a PVC, and the namespace it lives in, launch some kind of container that mounts it

  Args:
  ----
    pvc_name (str): The name of the PVC to launch a management container for
    namespace (str): The namespace the PVC is located in

  """
  pod_name = _find_pod_by_pvc(pvc_name)

  if pod_name != "" and pod_name != "pvc-manager":
    volume_name = _find_volume_by_pvc_and_pod(pod_name, namespace, pvc_name)
    _delete_pvc_manager_if_exists(pod_name, namespace)

    time.sleep(6)

    _patch_pod_with_devterm_container(pod_name, namespace, volume_name)
  else:
    _create_pvc_manager_pod(pvc_name, namespace)

    pod_name = "pvc-manager"
    volume_name = pvc_name

  time.sleep(5)

  _exec_shell_in_pod(pod_name, namespace, "pvc-manager")


def list_available_jobs() -> None:
  # TODO exclude the docs directory, include a list of valid arguments
  """Print all the jobs available to run in the jobs directory to the console"""
  jobs_dir = config.get_full_atils_dir("JOBS_DIR")

  root, dirs, files = next(os.walk(jobs_dir))
  for job in dirs:
    if job != "docs":
      jobconfig_path = os.path.join(jobs_dir, job, ".atils_jobconfig.json")
      if os.path.exists(jobconfig_path):
        with open(jobconfig_path) as file:
          jobconfig_data = json.load(file)
          description = jobconfig_data["short_description"]

        print(f"{job}:  {description}")


def run_job(job_name: str, args=None) -> None:
  """Given a job name and list of args, render the job template, then run the job
  Args:
    job_name (str): The name of the job to run. Must be a directory in the JOBS_DIR directory
    args (dict[str, str]): A dictionary representing arguments. Each key should correspond to a
    variable in a job template, with each value representing what should be filled in
  """
  rendered_job = _render_job(job_name, args)
  _launch_job(rendered_job)
  logging.info(f"Job {job_name} created")


def _clear_job_name(job_name: str, namespace: str) -> None:
  """We don't do a GenerateName for our jobs, so we need to make sure that the generated job name is available.
  So given a job name, and a namespace, delete the job, and then make sure it's deleted before letting us out
  """
  # Get all the jobs in the namespace, and then loop over them, looking for a matching name field
  # If found, we'll then delete the job, and wait for it to clear out
  v1 = client.BatchV1Api()
  for job in v1.list_namespaced_job(namespace).items:
    if job.metadata.name == job_name:
      # TODO Let's also delete all pods associated with the job
      # TODO the best way to do that is going to be to try and get a pod with all the matching labels,
      # so let's just refactor to not be afraid of error handling
      v1.delete_namespaced_job(name=job_name, namespace=namespace)
      # Wait until the job is deleted
      dots = itertools.cycle([".  ", ".. ", "..."])
      spinner = itertools.cycle(["-", "\\", "|", "/"])

      # TODO split out the waiting logic
      job = v1.read_namespaced_job(name=job_name, namespace=namespace)
      while job:
        try:
          job = v1.read_namespaced_job(name=job_name, namespace=namespace)
          print(
            f"Waiting for job {job_name} to be deleted{next(dots)} {next(spinner)}",
            end="\r",
          )
          time.sleep(0.2)
        except client.rest.ApiException as e:
          if e.status == 404:
            job = None
          else:
            raise e
      print("\n")
      logging.info(f"Job {job_name} deleted")
      return
  logging.info(f"No job named {job_name} found in namespace {namespace}")


def _create_pod_if_not_existent(
  pod_name: str,
  pod_manifest: dict[str, Any],
  namespace: str = "default",
) -> None:
  # TODO we can make this generic with a ton of arguments
  """Create a pod that we want to exec into (generally) if one doesn't exist. We're also going to wait for it
  to be ready

  Args:
  ----
    pod_name (str): The name of the pod we will want to create
    pod_manifest (dict[str, Any]): A dictionary representing a valid Kubernetes pod object
    namespace: The namespace this pod will be launched in. Defaults to "default"

  """
  try:
    api_instance = client.CoreV1Api()

    try:
      existing_pod = api_instance.read_namespaced_pod(
        name=pod_name,
        namespace=namespace,
      )

      # TODO right now, we might get kicked out of our pod if it's close to expiring. That's probably
      # not a big deal, but if so, fix it here
      if existing_pod:
        logging.info(
          f"There's already a {pod_name} pod! We're just gonna leave it",
        )
        return
    except ApiException as e:
      if e.status != 404:
        logging.exception(
          f"Error checking for existing pod 'postgres-manager': {e!s}",
        )
        exit(1)

    try:
      # Create the pod
      api_instance.create_namespaced_pod(namespace=namespace, body=pod_manifest)
      logging.info(f"Created pod '{pod_name}' in {namespace} namespace")

      # Now, let's wait for it to be ready
      pod_ready = False
      while not pod_ready:
        pod = api_instance.read_namespaced_pod(
          name=pod_name,
          namespace=namespace,
        )

        if pod.status.phase == "Running":
          for container in pod.status.container_statuses:
            if not container.ready:
              break
          else:
            print(f"Pod {pod_name} is ready.")
            return

        time.sleep(0.5)

    except client.rest.ApiException as e:
      logging.exception(f"Error creating pod '{pod_name}': {e!s}")

  except Exception as e:
    logging.exception(f"Error occurred while creating pod 'postgres-manager': {e!s}")


def _create_pvc_manager_pod(pvc_name: str, namespace: str) -> None:
  try:
    api_instance = client.CoreV1Api()

    # Check if a pod named "pvc-manager" already exists in the namespace
    try:
      existing_pod = api_instance.read_namespaced_pod(
        name="pvc-manager",
        namespace=namespace,
      )
      if existing_pod:
        # Delete the existing "pvc-manager" pod
        api_instance.delete_namespaced_pod(
          name="pvc-manager",
          namespace=namespace,
          grace_period_seconds=0,
        )
        logging.info(
          f"Deleted existing pod 'pvc-manager' in namespace '{namespace}'",
        )
        time.sleep(5)  # Wait for the pod to be deleted
    except ApiException as e:
      if e.status != 404:
        logging.exception(
          f"Error checking/deleting existing pod 'pvc-manager': {e!s}",
        )

    # Define the pod manifest
    pod_manifest = {
      "apiVersion": "v1",
      "kind": "Pod",
      "metadata": {"name": "pvc-manager"},
      "spec": {
        "containers": [
          {
            "name": "pvc-manager",
            "image": "aidanhilt/atils-devterm",
            "command": ["/bin/sh"],
            "args": [
              "-c",
              "sleep 1800",
            ],  # Sleep for 30 minutes (1800 seconds)
            "volumeMounts": [
              {"name": "pvc", "mountPath": f"/root/{pvc_name}"},
            ],
          },
        ],
        "volumes": [
          {"name": "pvc", "persistentVolumeClaim": {"claimName": pvc_name}},
        ],
        "restartPolicy": "Never",
      },
    }

    try:
      # Create the pod
      api_instance.create_namespaced_pod(namespace=namespace, body=pod_manifest)
      logging.info(
        f"Created pod 'pvc-manager' with PVC '{pvc_name}' mounted at '/root/{pvc_name}'",
      )
    except client.rest.ApiException as e:
      logging.exception(f"Error creating pod 'pvc-manager': {e!s}")

  except Exception as e:
    logging.exception(f"Error occurred while creating pod: {e!s}")


def _delete_pvc_manager_if_exists(pod_name: str, namespace: str) -> None:
  try:
    api_instance = client.CoreV1Api()

    try:
      # Get the pod details
      pod = api_instance.read_namespaced_pod(name=pod_name, namespace=namespace)

      # Check if the pod has ephemeral containers
      if pod.spec.ephemeral_containers:
        # Find the index of the "pvc-manager" ephemeral container
        container_index = next(
          (
            index
            for index, container in enumerate(pod.spec.ephemeral_containers)
            if container.name == "pvc-manager"
          ),
          None,
        )

        if container_index is not None:
          # Remove the "pvc-manager" ephemeral container from the pod spec
          pod.spec.ephemeral_containers.pop(container_index)

          # Patch the pod to update the ephemeral containers
          api_instance.patch_namespaced_pod(
            name=pod_name,
            namespace=namespace,
            body=pod,
          )
          logging.devterm(
            f"Deleted ephemeral container 'pvc-manager' from pod '{pod_name}'",
          )
        else:
          logging.devterm(
            f"Ephemeral container 'pvc-manager' not found in pod '{pod_name}'",
          )
      else:
        logging.devterm(f"No ephemeral containers found in pod '{pod_name}'")

    except ApiException as e:
      if e.status == 404:
        logging.warning(
          f"Pod '{pod_name}' not found in namespace '{namespace}'",
        )
      else:
        logging.exception(
          f"Error deleting ephemeral container from pod '{pod_name}': {e!s}",
        )

  except Exception as e:
    logging.exception(f"Error occurred while deleting ephemeral container: {e!s}")


def _ensure_access_exists(
  service_account_name: str,
  service_account_namespace: str,
  role_name: str,
  cluster_role: bool = False,
) -> None:
  """Given a service account, ensure that it's bound to either a Role, or a ClusterRole based on the cluster_role arg

  Args:
  ----
    service_account_name (str): The name of the service account we want to use for our job
    service_account_namespace (str): The namespace our chosen service account lives in
    role_name (str): The name of the role our service account needs to be bound to. If cluster_role is set to True, this is a ClusterRole,
    and it is a namespaced role otherwise

  """
  # Create API clients
  v1 = client.CoreV1Api()
  rbac_v1 = client.RbacAuthorizationV1Api()

  # 1. Check if the service account exists
  try:
    v1.read_namespaced_service_account(
      name=service_account_name,
      namespace=service_account_namespace,
    )
  except ApiException as e:
    if e.status == 404:
      logging.exception(
        f"Service account {service_account_name} does not exist in namespace {service_account_namespace}.",
      )
      exit(1)
    else:
      logging.exception(f"Error checking service account: {e}")
      exit(1)

  # 2. Check if the role/cluster role exists
  try:
    if cluster_role:
      rbac_v1.read_cluster_role(name=role_name)
    else:
      rbac_v1.read_namespaced_role(
        name=role_name,
        namespace=service_account_namespace,
      )
  except ApiException as e:
    if e.status == 404:
      logging.exception(
        f"{'ClusterRole' if cluster_role else 'Role'} {role_name} does not exist.",
      )
      return
    else:
      logging.exception(
        f"Error checking {'cluster role' if cluster_role else 'role'}: {e}",
      )
      return

  # 3. Check and create role binding if necessary
  binding_name = f"{service_account_name}-{role_name}"
  try:
    if cluster_role:
      rbac_v1.read_cluster_role_binding(name=binding_name)
    else:
      rbac_v1.read_namespaced_role_binding(
        name=binding_name,
        namespace=service_account_namespace,
      )
    print(
      f"{'ClusterRoleBinding' if cluster_role else 'RoleBinding'} {binding_name} already exists.",
    )
  except ApiException as e:
    if e.status == 404:
      # Create the binding
      if cluster_role:
        body = client.V1ClusterRoleBinding(
          metadata=client.V1ObjectMeta(name=binding_name),
          subjects=[
            client.V1Subject(
              kind="ServiceAccount",
              name=service_account_name,
              namespace=service_account_namespace,
            ),
          ],
          role_ref=client.V1RoleRef(
            kind="ClusterRole",
            name=role_name,
            api_group="rbac.authorization.k8s.io",
          ),
        )
        rbac_v1.create_cluster_role_binding(body=body)
      else:
        body = client.V1RoleBinding(
          metadata=client.V1ObjectMeta(
            name=binding_name,
            namespace=service_account_namespace,
          ),
          subjects=[
            client.V1Subject(
              kind="ServiceAccount",
              name=service_account_name,
              namespace=service_account_namespace,
            ),
          ],
          role_ref=client.V1RoleRef(
            kind="Role",
            name=role_name,
            api_group="rbac.authorization.k8s.io",
          ),
        )
        rbac_v1.create_namespaced_role_binding(
          namespace=service_account_namespace,
          body=body,
        )
      print(
        f"Created {'ClusterRoleBinding' if cluster_role else 'RoleBinding'} {binding_name}",
      )
    else:
      print(
        f"Error checking {'cluster role binding' if cluster_role else 'role binding'}: {e}",
      )


def _exec_shell_in_pod(
  pod_name: str,
  namespace: str,
  container_name: str,
  exec_command: list[str] = ["/bin/zsh"],
  delete: bool = False,
) -> None:
  """Exec into a given container in a given pod, in a given namespace. This will assume
  that the container has zsh installed

  Args:
  ----
    pod_name (str): The name of the pod to exec into
    namespace (str): The namespace the pod is located in
    container_name (str): The name of the container to exec into

  """
  api_client = client.ApiClient()
  api_instance = client.CoreV1Api(api_client)

  resp = stream(
    api_instance.connect_get_namespaced_pod_exec,
    pod_name,
    namespace,
    command=exec_command,
    container=container_name,
    stderr=True,
    stdin=True,
    stdout=True,
    tty=True,
    _preload_content=False,
  )

  t = Thread(target=_read, args=[resp])

  # change tty mode to be able to work with escape characters
  stdin_fd = sys.stdin.fileno()
  old_settings = termios.tcgetattr(stdin_fd)
  try:
    tty.setraw(stdin_fd)
    t.start()
    while resp.is_open():
      data = resp.read_stdout(10)
      if resp.is_open():
        if len(data or "") > 0:
          sys.stdout.write(data)
          sys.stdout.flush()
  finally:
    # reset tty
    if delete:
      api_instance.delete_namespaced_pod(pod_name, namespace)
    print("\033c")
    termios.tcsetattr(stdin_fd, termios.TCSADRAIN, old_settings)
    print("press enter")


def _fill_out_jobconfig_args_from_command_line(
  jobconfig_args: dict,
  command_line_args: dict,
) -> dict[str, str]:
  """Fill out the jobconfig_args dictionary with the c.

  Args:
  ----
    jobconfig_args (dict): The dictionary to fill out
    command_line_args (dict): The dictionary containing the command line arguments

  Returns:
  -------
    dict: The updated jobconfig_args dictionary

  """
  for key, value in command_line_args.items():
    jobconfig_args[key] = value
  return jobconfig_args


def _find_pod_by_pvc(pvc_name: str) -> str:
  """Given the name of a PVC, find the name of a pod it is attached to. If no pod is attached, return an empty string

  Args:
  ----
    pvc_name (str): The name of the PVC to search for

  Returns:
  -------
    str: The name of the pod the PVC is attached to, or an empty string if no pod is attached.

  """
  try:
    v1 = client.CoreV1Api()

    # List all pods in all namespaces
    pods: List[client.V1Pod] = v1.list_pod_for_all_namespaces().items

    for pod in pods:
      for volume in pod.spec.volumes:
        if volume.persistent_volume_claim and volume.persistent_volume_claim.claim_name == pvc_name:
          return pod.metadata.name

    return ""

  except Exception as e:
    print(f"Error occurred while searching for pod: {e!s}")
    return ""


def _find_volume_by_pvc_and_pod(pod_name: str, namespace: str, pvc_name: str) -> str:
  """Given a pvc name, and the name of a pod, find the name the volume was given, for mounting purposes.
  We assume there's a volume here, so fail if nothing is found

  Args:
  ----
    pod_name (str): The name of the pod to search for
    namespace (str): The namespace the pod is in
    pvc_name (str): The name of the PVC to search for

  Returns:
  -------
    str: The name of the volume that mounts our pvc

  """
  try:
    v1 = client.CoreV1Api()
    pod = v1.read_namespaced_pod(name=pod_name, namespace=namespace)

    for volume in pod.spec.volumes:
      if volume.persistent_volume_claim and volume.persistent_volume_claim.claim_name == pvc_name:
        return volume.name

    logging.error(
      f"No volume found using PVC '{pvc_name}' in pod '{pod_name}' in namespace '{namespace}'",
    )
    sys.exit(1)

  except Exception as e:
    logging.exception(f"Error occurred while searching for volume: {e!s}")
    sys.exit(1)


def _get_arguments_from_jobconfig(job_name: str) -> dict[str, str]:
  """Get the arguments from the jobconfig file.

  Args:
  ----
    job_name (str): The name of the job to get the arguments for

  Returns:
  -------
    dict[str, str]: A dictionary containing the arguments

  """
  try:
    jobs_dir = config.get_full_atils_dir("JOBS_DIR")
    if os.path.exists(os.path.join(jobs_dir, job_name, ".atils_jobconfig.json")):
      jobconfig_file = os.path.join(jobs_dir, job_name, ".atils_jobconfig.json")
    else:
      logging.error(f"No jobconfig found for job {job_name}, exiting")
      sys.exit(1)

    jobconfig_args = {}

    # Read the jobconfig file
    with open(jobconfig_file) as file:
      jobconfig = json.load(file)

      if "args" in jobconfig.keys():
        for arg in jobconfig["args"]:
          if "default" in arg.keys():
            jobconfig_args[arg["name"]] = arg["default"]
          else:
            jobconfig_args[arg["name"]] = ""

    return jobconfig_args

  except Exception as e:
    logging.exception(
      f"Error occurred while getting arguments from jobconfig: {e!s}",
    )
    sys.exit(1)


def _get_missing_arguments_interactive(
  missing_args_dict: dict[str, str],
) -> dict[str, str]:
  """Get missing job arguments from the user.

  Args:
  ----
    missing_args_dict (dict[str, str]): A dictionary containing the missing arguments

  Returns:
  -------
    dict[str, str]: A dictionary containing the updated missing arguments

  """
  try:
    for arg_name, arg_value in missing_args_dict.items():
      if arg_value == "":
        arg_value = input(f"Enter value for {arg_name}: ")
        missing_args_dict[arg_name] = arg_value

    return missing_args_dict

  except Exception as e:
    logging.exception(f"Error occurred while getting missing arguments: {e!s}")
    sys.exit(1)


def _launch_job(job_dict):
  job_name = job_dict["metadata"]["name"]

  if "namespace" in job_dict["metadata"]:
    namespace = job_dict["metadata"]["namespace"]
  else:
    _, active_context = k8s_config.list_kube_config_contexts()
    if "namespace" in active_context["context"]:
      namespace = active_context["context"]["namespace"]
    else:
      namespace = "default"
    job_dict["metadata"]["namespace"] = namespace

  _clear_job_name(job_name, namespace)

  k8s_client = client.ApiClient()
  utils.create_from_dict(k8s_client, job_dict)


def _patch_pod_with_devterm_container(
  pod_name: str,
  namespace: str,
  volume_name: str,
) -> None:
  """Patch a pod with an ephemeral container, running our devterm image. This then mounts a PVC in the home directory,
  to view and modify any files

  Args:
  ----
    pod_name (str): The name of the pod to patch
    namespace (str): The namespace the pod to patch lives in
    volume_name (str): The name of the volume to mount in the pod

  """
  try:
    api_instance = client.CoreV1Api()

    # Define the ephemeral container
    ephemeral_container = {
      "name": "pvc-manager",
      "image": "aidanhilt/atils-devterm",
      "command": ["/bin/sh"],
      "args": ["-c", "sleep 1800"],  # Sleep for 30 minutes (1800 seconds)
      "volumeMounts": [
        {"name": volume_name, "mountPath": f"/root/{volume_name}"},
      ],
    }

    body = {"spec": {"ephemeralContainers": [ephemeral_container]}}

    try:
      # Patch the pod with the ephemeral container
      api_instance.patch_namespaced_pod_ephemeralcontainers(
        name=pod_name,
        namespace=namespace,
        body=body,
      )

      logging.devterm(
        f"Successfully patched pod '{pod_name}' with ephemeral container",
      )
    except Exception as e:
      logging.exception(f"Error patching pod '{pod_name}': {e!s}")

  except Exception as e:
    logging.exception(f"Error occurred while patching pod: {e!s}")


def _read(resp):
  """Redirect the terminal input to the stream, and read the response from the stream.
  This is used to read the response from the stream when we are running a job.

  Args:
  ----
    resp (stream): The stream object to read from.

  Returns:
  -------
    str: The response from the stream.

  """
  while resp.is_open():
    char = sys.stdin.read(1)
    resp.update()
    if resp.is_open():
      resp.write_stdin(char)


def _render_job(job_name: str, args: dict[str, str]) -> str:
  """Given the name of a job, that is the same as a directory in the JOBS_DIR directory,
  render the template with the arguments provided, and return it
  Args:
    job_name (str): The name a job in the JOBS_DIR directory
    args (dict[str, str]): A dictionary representing arguments. Each key should correspond to a
    variable in a job template, with each value representing what should be filled in
  Returns:
    str: The contents of the template file, rendered with the values of args
  """
  jobs_dir = config.get_full_atils_dir("JOBS_DIR")
  if os.path.exists(os.path.join(jobs_dir, job_name, "job.yaml")):
    rendered_job = template_utils.template_external_file(
      os.path.join(jobs_dir, job_name, "job.yaml"),
      args,
    )
    return yaml.safe_load(rendered_job)

  else:
    logging.error(f'Job "{job_name}" was not found')
    exit(1)



import argparse
import asyncio
import logging
import os
import subprocess
import sys
import time

import watchfiles

from atils.common import config
from atils.common.settings import settings

logging.basicConfig(level=config.get_logging_level())  # type: ignore
logging.getLogger("asyncio").setLevel(logging.WARNING)
watchfiles.main.logger.setLevel(logging.WARNING)


def main(args: str):
  parser: argparse.ArgumentParser = argparse.ArgumentParser()
  subparsers = parser.add_subparsers(
    help="Select a subcommand",
    dest="subparser_name",
  )

  auto_deploy_parser = subparsers.add_parser(
    "auto-deploy",
    help="Watch a helm chart for changes, and deploy automatically",
  )

  auto_deploy_parser.add_argument(
    "--chart-name",
    "-cn",
    type=str,
    help="The name of the helm chart to watch",
    required=True,
  )

  if len(args) == 0:
    parser.print_help(sys.stderr)
    sys.exit(1)

  arguments: argparse.Namespace = parser.parse_args(args)

  # Let's explain this a little bit. We run the auto_deploy_helm_chart async, so that when we get a
  # KeyboardInterrupt (i.e. Ctrl+C ), we can uninstall the chart automatically.
  if arguments.subparser_name == "auto-deploy":
    try:
      asyncio.run(auto_deploy_helm_chart(arguments.chart_name))
    except KeyboardInterrupt:
      subprocess.run(["helm", "uninstall", arguments.chart_name], check=False)


async def auto_deploy_helm_chart(chart_name: str) -> None:
  """Automatically deploy a helm chart in the HELM_CHARTS_DIR directory. After the initial deployment,
  it then reinstalls the chart whenever an update is made to the chart, or to the values file located at
  ~/.atils/helm-values/<chart-name>.yaml

  Args:
  ----
    chart_name (str): The name of the helm chart to deploy.

  """
  helm_chart_dir = os.path.join(
    settings.install_dir,
    settings.helm_charts_dir,
    chart_name,
  )

  # Check if there is a directory at the given path, that has a Chart.yaml file in it
  if os.path.isdir(helm_chart_dir) and os.path.isfile(
    os.path.join(helm_chart_dir, "Chart.yaml"),
  ):
    print(f"Watching and auto-installing {chart_name}...")
    _install_helm_chart(chart_name, helm_chart_dir)
    # This is where we watch for changes to the chart and values file. Those specific functions then
    # handle calling the install function
    await asyncio.gather(
      _watch_chart_path(chart_name),
      _watch_values_path(chart_name),
    )

  else:
    logging.error(f"No chart found at the given path: {helm_chart_dir}")
    exit(1)


def _install_helm_chart(chart_name: str, helm_chart_dir: str) -> None:
  """Automatically installs a helm chart located at helm_chart_dir. If there is a file at
  ~/.atils/helm-values/<chart_name>.yaml, this functions uses that as the values.yaml file

  Args:
  ----
    chart_name (str): The name of the helm chart to install, which is used as the release name
    helm_chart_dir (str): The path to the helm chart to install

  """
  values_file = os.path.join(
    settings.config_directory,
    "helm-values",
    f"{chart_name}.yaml",
  )
  if os.path.isfile(values_file):
    subprocess.run(
      [
        "helm",
        "upgrade",
        "--install",
        chart_name,
        helm_chart_dir,
        "-f",
        values_file,
      ],
      check=False,
    )
  else:
    subprocess.run(["helm", "upgrade", "--install", chart_name, helm_chart_dir], check=False)


async def _watch_chart_path(chart_name: str) -> None:
  """Watches a local helm chart for changes, and calls _install_helm_chart whenever an update is detected

  Args:
  ----
    chart_name (str): The name of the chart. This will be the name of a directory in the HELM_CHARTS_DIR

  """
  helm_chart_dir = os.path.join(
    settings.install_dir,
    settings.helm_charts_dir,
    chart_name,
  )

  async for change in watchfiles.awatch(helm_chart_dir):
    _install_helm_chart(chart_name, helm_chart_dir)


async def _watch_values_path(chart_name: str) -> None:
  """Watches the values file in ~/.atils/helm-values/<chart_name>.yaml for changes, and calls _install_helm_chart
  whenever an update is detected

  Args:
  ----
    chart_name (str): The name of the chart. This will be the name of a directory in the HELM_CHARTS_DIR

  """
  helm_chart_dir = os.path.join(
    settings.install_dir,
    settings.helm_charts_dir,
    chart_name,
  )

  values_file = os.path.join(
    settings.config_directory,
    "helm-values",
    f"{chart_name}.yaml",
  )
  # TODO we may want to clean this up, I don't think we need to loop in this function, when we're calling it in a loop
  while True:
    if os.path.isfile(values_file):
      async for change in watchfiles.awatch(values_file):
        _install_helm_chart(chart_name, helm_chart_dir)
    else:
      time.sleep(5)
